/*

File: proxyDaemon.c

*/

#include <sys/types.h>
#include <sys/event.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <syslog.h>
#include <libgen.h>
#include <launch.h>
#include <asl.h>


// Commands for proxydaemon to reduce plain text over the pipes 
static const char READ_ALL[] = "1000";  // Read and parse all data
static const char SET_APD[]  = "1001";  // AutoProxyDiscovery(boolean)
static const char SET_APCE[] = "1002";  // AutoProxyConfigurationEnable(boolean)
static const char SET_APCU[] = "1003";  // AutoProxyConfigurationURL(String)
static const char SET_WPE[]  = "1004";  // WebProxyEnable(boolean)
static const char SET_WP[]   = "1005";  // WebProxy(String, int, booean)
static const char SET_SWPE[] = "1006";  // SecureWebProxyEnable(boolean)
static const char SET_SWP[]  = "1007";  // SecureWebProxy(String, int, booean)
static const char SET_SPE[]  = "1008";  // SocksProxyEnable(boolean)
static const char SET_SP[]   = "1009";  // SocksProxy(String, int, booean)
static const char SET_NPF[]  = "1010";  // NoProxyFor(String)

FILE *the_stream = NULL;
int fd;
aslclient asl = NULL;
aslmsg log_msg = NULL;

int pipe_bidi(char *cmd, char *data, char *result, int log)
{
    FILE *ptr;
    char buf[BUFSIZ];

    if ((ptr = popen(cmd, "r+")) != NULL)
    {
	if (NULL != data) {
	    (void) fprintf(ptr, "%s", data);
	    fflush(ptr);
	}
    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "Wrote:%s <%s>", cmd, data);
	strcpy(result, "");
	while(NULL != fgets(buf, BUFSIZ, ptr))
	{
	    strcat(result, buf);
	    if (strstr(buf, "}") != NULL) break;
	}
    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "Read <%s>", result);
	return (pclose(ptr));
    }
    return 88;
}

void SigPipeHandler(int s);

void SigPipeHandler(int s)
{
	fflush(the_stream);
	fclose(the_stream);
	close(fd);
	the_stream = NULL;  
	asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "SigPipe %d", s);
}


int main(int argc, char *argv[])
{
    struct sockaddr_storage ss;
    socklen_t slen = sizeof(ss);
    struct kevent kev_init;
    struct kevent kev_listener;
    launch_data_t sockets_dict, checkin_response;
    launch_data_t checkin_request;
    launch_data_t listening_fd_array;
    size_t i;
    int kq;
    int retval = EXIT_SUCCESS;
    int log = FALSE;

    /*
     * Create a new ASL log
     *
     */	 
    asl = asl_open("proxyDaemon", "Daemon", ASL_OPT_STDERR);
    log_msg = asl_new(ASL_TYPE_MSG);
    asl_set(log_msg, ASL_KEY_SENDER, "proxyDaemon");
    
	/*
    * Create a new kernel event queue
    * that we'll use for our notification.
    *
    */
    if (-1 == (kq = kqueue())) {
		retval = 10;
		goto done;
    }

    /*
    * Register ourselves with launchd.
    *
    */
    if ((checkin_request = launch_data_new_string(LAUNCH_KEY_CHECKIN)) == NULL) {
		retval = 1;
		goto done;
    }

    if ((checkin_response = launch_msg(checkin_request)) == NULL) {
		retval = 2;
		goto done;
    }

    if (LAUNCH_DATA_ERRNO == launch_data_get_type(checkin_response)) {
		errno = launch_data_get_errno(checkin_response);
		retval = 3;
		goto done;
    }

    launch_data_t the_label = launch_data_dict_lookup(checkin_response, LAUNCH_JOBKEY_LABEL);
    if (NULL == the_label) {
		retval = 4;
		goto done;
    }


    /*
    * Retrieve the dictionary of Socket entries in the config file
    */
    sockets_dict = launch_data_dict_lookup(checkin_response, LAUNCH_JOBKEY_SOCKETS);
    if (NULL == sockets_dict) {
		retval = 5;
		goto done;
    }

    if (launch_data_dict_get_count(sockets_dict) > 1) {
	// We only use the first socket if multiple are defined.
    }

    /*
    * Get the dictionary value from the key "MyListenerSocket",
    * as defined in the plist file.
    */
    listening_fd_array = launch_data_dict_lookup(sockets_dict, "MyListenerSocket");
    if (NULL == listening_fd_array) {
		retval = 6;
		goto done;
    }

    /*
    * Initialize a new kernel event.  This will trigger when
    * a connection occurs on our listener socket.
    *
    */
    for (i = 0; i < launch_data_array_get_count(listening_fd_array); i++) {
		launch_data_t this_listening_fd = launch_data_array_get_index(listening_fd_array, i);

		EV_SET(&kev_init, launch_data_get_fd(this_listening_fd), EVFILT_READ, EV_ADD, 0, 0, NULL);
		if (kevent(kq, &kev_init, 1, NULL, 0, NULL) == -1) {
	    	retval = 7;
	    	goto done;
		}
    }

    launch_data_free(checkin_response);


    char tbuf[512];
    int ret;

    for (;;) {	// Keep reading until exit command
		retval = 9;
		if (the_stream == NULL)	// If we have open stream, keep using
		{
			/*
			*
			* Get the next event from the kernel event queue.
			*
			*/
			if ((fd = kevent(kq, NULL, 0, &kev_listener, 1, NULL)) == -1) {
			retval = 8;
			goto done;
			} else if (fd == 0) {
			retval = 10; //EXIT_SUCCESS;
			goto done;
			}

			/*
			*
			* Accept an incoming connection.
			*
			*/
			if ((fd = accept(kev_listener.ident, (struct sockaddr *)&ss, &slen)) == -1) {
			goto done;//   continue; /* this isn't fatal */
			}

			/*
			* Open the stream and message the client.
			*/
			the_stream = fdopen(fd, "r+");
		    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "Connection established");
			signal(SIGPIPE, SigPipeHandler);
		}
		char command[600];  // Bigger than read to prevent Buffer overrun
		char newCommand[600];  // Bigger than read to prevent Buffer overrun
		char theName[512];

		if (the_stream) {
			if(fgets(tbuf, 512, the_stream) != NULL)
			{
				strcpy(command, tbuf);
			    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "Command:%s", command);
			}
			if (command == strstr(command, "<ServiceName> ")) {
			       	char *serviceName = command + 14;
       				char *xx = strstr(serviceName, " ");
       				(*xx) = 0;
       				strcpy(theName, serviceName);
       				strcpy(newCommand,xx + 1);
				strcpy(command, newCommand);
			} else {
				if(argc > 1) {
					strcpy(theName, argv[1]);
				}
			}

			if (command == strstr(command, "<Exit>"))
			{
				fclose(the_stream);
				close(fd);
				the_stream = NULL;
			    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "Exiting...");
				return 0;
			}

			if (command == strstr(command, "<End>"))
			{
				fclose(the_stream);
				close(fd);
				the_stream = NULL;
				continue;
			}

			if (command == strstr(command, "<Log>"))
			{
				log = TRUE;
			    asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "Logging on...");
				fprintf(the_stream, "%s", "\n<Ready>\n\n");
				fflush(the_stream);
				continue;
			}

			if (command == strstr(command, "<NoLog>"))
			{
				log = FALSE;
			    asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "Logging off...");
				fprintf(the_stream, "%s", "\n<Ready>\n\n");
				fflush(the_stream);
				continue;
			}
			char result[1024];

			// Read IPv4 dictionary to Service GUID
			retval = 180 + pipe_bidi("scutil", "open\nget State:/Network/Global/IPv4\nd.show\n", result, log);
			if (retval != 180) goto done;

			/*  Result is like this
				<dictionary> {
					PrimaryInterface : en0
					PrimaryService : 8025D2F9-FFFB-4A5A-B167-E6589CF8BB9B
					Router : 192.168.1.1
				}
			*/

			char *off;
			char *end;

			// Find start of Primary service
			if (NULL == (off = strstr(result, "PrimaryService")))
				return 77;
			off += 17;  // Skip over it and " : "

			if (NULL == (end = strstr(off, "\n")))
				return 66;
			*end = 0;	// Null out the newline

			char dataToSend[1024];

			if (strlen(off) > 10 ) { // If it looks like a GUID, take it rather than current value for theName
				// Now use the GUID to find Service's name
				strcpy(dataToSend,  "open\nget Setup:/Network/Service/");
				strcat(dataToSend,  off);
				strcat(dataToSend,  "\nd.show\n");
				result[0] = 0;

				retval = 190 + pipe_bidi("scutil", dataToSend, result, log);
				if (retval != 190) goto done;

				/*
					<dictionary> {
						UserDefinedName : Ethernet Home
					}
				*/

				// Extract pointer to the name
				if (NULL == (off = strstr(result, "UserDefinedName")))
					return 55;
				off += 18;

				if (NULL == (end = strstr(off, "\n")))
					return 44;
				*end = 0;

				strcpy(theName, off);
			}
			/*
			*
			* Now begin processing commands
			*
			*/

			//  SET_APD[]  = "1001";  // AutoProxyDiscovery(boolean)
			if (command == strstr(command, SET_APD))
			{
				result[0] = 0;
				char *offSet = command + 5;
				if (offSet == strstr(offSet, "on"))
					offSet[2] = 0;
				else
				if (offSet == strstr(offSet, "off"))
					offSet[3] = 0;
				else
				{
					fflush(the_stream);
					retval = 177;
					goto done;
				}
				strcpy(dataToSend,  "networksetup  -setproxyautodiscovery \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\" ");
				strcat(dataToSend,  offSet);
				retval = 170 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 170) goto done;
			}
			else
			//  SET_APCE[] = "1002";  // AutoProxyConfigurationEnable(boolean)
			if (command == strstr(command, SET_APCE))
			{
				result[0] = 0;
				char *offSet = command + 5;
				if (offSet == strstr(offSet, "on"))
					offSet[2] = 0;
				else
				if (offSet == strstr(offSet, "off"))
					offSet[3] = 0;
				else
				{
					fflush(the_stream);
					retval = 175;
					goto done;
				}
				strcpy(dataToSend,  "networksetup  -setautoproxystate \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\" ");
				strcat(dataToSend,  offSet);
				retval = 170 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 170) goto done;
			}
			else
			//  SET_APCU[] = "1003";  // AutoProxyConfigurationURL(String)
			if (command == strstr(command, SET_APCU))
			{
				result[0] = 0;
				char *offSet = command + 5;
				if (NULL != (end = strstr(offSet, "\n")))
					*end = 0;
				else
				{
					fflush(the_stream);
					retval = 175;
					goto done;
				}
				strcpy(dataToSend,  "networksetup  -setautoproxystate \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\" \"");
				strcat(dataToSend,  offSet);
				strcat(dataToSend,  "\"");
				retval = 170 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 170) goto done;
			}
			else
			//  SET_WPE[]  = "1004";  // WebProxyEnable(boolean)
			if (command == strstr(command, SET_WPE))
			{
				result[0] = 0;
				char *offSet = command + 5;
				if (offSet == strstr(offSet, "on"))
					offSet[2] = 0;
				else
				if (offSet == strstr(offSet, "off"))
					offSet[3] = 0;
				else
				{
					fflush(the_stream);
					retval = 174;
					goto done;
				}
				strcpy(dataToSend,  "networksetup  -setwebproxystate \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\" ");
				strcat(dataToSend,  offSet);
				retval = 170 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 170) goto done;
			}
			else
			//  SET_WP[]   = "1005";  // WebProxy(String, int, booean)
			//  SET_SWP[]  = "1007";  // SecureWebProxy(String, int, booean)
			//  SET_SP[]   = "1009";  // SocksProxy(String, int, booean)
			if ((command == strstr(command, SET_WP)) ||
				 (command == strstr(command, SET_SWP)) ||
				 (command == strstr(command, SET_SP)))
			{
				result[0] = 0;
				// Quick syntax check
				int i = 0;
				char *token[100];  // Big for safety
				token[i] = strtok(command, "|"); 
				while(token[i]!= NULL) {   //ensure a pointer was found
				    i++;
				    token[i] = strtok(NULL, "|"); //continue to tokenize the string
				}
				if (i < 3)
				{
					fflush(the_stream);
					retval = 172;
					goto done;
				}
				if (command == strstr(command, SET_WP))
					strcpy(dataToSend,  "networksetup  -setwebproxy \"");
				else
				if (command == strstr(command, SET_SWP))
					strcpy(dataToSend,  "networksetup  -setsecurewebproxy \"");
				else
					strcpy(dataToSend,  "networksetup  -setsocksfirewallproxy \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\" \"");
				strcat(dataToSend,  token[1]);
				strcat(dataToSend,  "\" \"");
				strcat(dataToSend,  token[2]);
				strcat(dataToSend,  "\"");
				if (i > 3)
				{
					strcat(dataToSend,  " ");
					strcat(dataToSend,  token[3]);
					if (i == 6)
					{
						strcat(dataToSend,  " \"");
						strcat(dataToSend,  token[4]);
						strcat(dataToSend,  "\" \"");
						strcat(dataToSend,  token[5]);
						strcat(dataToSend,  "\"");
					}
				}
				retval = 200 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 200) goto done;
			}
			else
			//  SET_SWPE[] = "1006";  // SecureWebProxyEnable(boolean)
			if (command == strstr(command, SET_SWPE))
			{
				result[0] = 0;
				char *offSet = command + 5;
				if (offSet == strstr(offSet, "on"))
					offSet[2] = 0;
				else
				if (offSet == strstr(offSet, "off"))
					offSet[3] = 0;
				else
				{
					fflush(the_stream);
					retval = 173;
					goto done;
				}
				strcpy(dataToSend,  "networksetup  -setsecurewebproxystate \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\" ");
				strcat(dataToSend,  offSet);
				retval = 170 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 170) goto done;
			}
			else
			//  SET_SPE[]  = "1008";  // SocksProxyEnable(boolean)
			if (command == strstr(command, SET_SPE))
			{
				result[0] = 0;
				char *offSet = command + 5;
				if (offSet == strstr(offSet, "on"))
					offSet[2] = 0;
				else
				if (offSet == strstr(offSet, "off"))
					offSet[3] = 0;
				else
				{
					fflush(the_stream);
					retval = 176;
					goto done;
				}
				strcpy(dataToSend,  "networksetup  -setsocksfirewallproxystate \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\" ");
				strcat(dataToSend,  offSet);
				retval = 170 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 170) goto done;
			}
			else
			//  SET_NPF[]  = "1010";  // NoProxyFor(String)
			if (command == strstr(command, SET_NPF))
			{
				result[0] = 0;
				int i = 0;
				char *token[100];  // Big for safety
				token[i] = strtok(command, "|"); 
				while(token[i]!= NULL) {   //ensure a pointer was found
				    i++;
				    token[i] = strtok(NULL, "|"); //continue to tokenize the string
				}
				strcpy(dataToSend,  "networksetup  -setproxybypassdomains \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\"");
				
			    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "I value is:%d", i);
			    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "%s", dataToSend);
				if (i == 1)  // none provided
				{
				   strcat(dataToSend, " \" \"");
				}
				else
				{
				    int j = 1;
				    while (j < i)
				    {
					strcat(dataToSend,  " \"");
					strcat(dataToSend,  token[j++]);
					strcat(dataToSend,  "\"");
				    }
				}
			    if (log == TRUE) asl_log(asl, log_msg, ASL_LEVEL_NOTICE, "%s", dataToSend);
				retval = 200 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 200) goto done;
			}
			else
			//  READ_ALL[] = "1000";  // Read and parse all data
			if (command == strstr(command, READ_ALL))
			{
				fprintf(the_stream, "Data: Network Service Name\n%s\n", theName);
				fflush(the_stream);

				result[0] = 0;
				strcpy(dataToSend,  "networksetup  -getproxyautodiscovery \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\"");
				retval = 170 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 170) goto done;
				fprintf(the_stream, "Data: Auto Discovery\n%s", result);
				fflush(the_stream);

				//fprintf(the_stream, "%s\n", "1"); fflush(the_stream);

				result[0] = 0;
				strcpy(dataToSend,  "networksetup  -getautoproxyurl \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\"");
				retval = 120 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 120) goto done;
				fprintf(the_stream, "Data: Automatic Proxy Configuration\n%s", result);
				fflush(the_stream);

				result[0] = 0;
				strcpy(dataToSend,  "networksetup  -getwebproxy \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\"");
				retval = 30 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 30) goto done;
				fprintf(the_stream, "Data: Web Proxy\n%s", result);
				fflush(the_stream);

				result[0] = 0;
				strcpy(dataToSend,  "networksetup  -getsecurewebproxy \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\"");
				retval = 30 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 30) goto done;
				fprintf(the_stream, "Data: Secure Web Proxy\n%s", result);
				fflush(the_stream);

				result[0] = 0;
				strcpy(dataToSend,  "networksetup  -getsocksfirewallproxy \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\"");
				retval = 30 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 30) goto done;
				fprintf(the_stream, "Data: Socks Proxy\n%s", result);
				fflush(the_stream);

				result[0] = 0;
				strcpy(dataToSend,  "networksetup  -getproxybypassdomains \"");
				strcat(dataToSend,  theName);
				strcat(dataToSend,  "\"");
				retval = 30 + pipe_bidi(dataToSend, NULL, result, log);
				if (retval != 30) goto done;
				fprintf(the_stream, "Data: No Proxy For\n%s", result);
				fflush(the_stream);
			}
			else
			{
				fclose(the_stream);
				close(fd);
				the_stream = NULL;
				continue;
			}
			fprintf(the_stream, "%s", "\n<Ready>\n\n");
			fflush(the_stream);
		}
    }
done:
	return retval;


}
